<?php

namespace Drupal\cl_server\Controller;

use Drupal\cl_components\Component\Component;
use Drupal\cl_components\Component\ComponentDiscovery;
use Drupal\cl_components\Exception\ComponentNotFoundException;
use Drupal\cl_components\Exception\TemplateNotFoundException;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\PageCache\ResponsePolicy\KillSwitch;
use Drupal\Core\Render\RendererInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

/**
 * Provides an endpoint for Storybook to query.
 *
 * @see https://github.com/storybookjs/storybook/tree/next/app/server
 */
class ServerEndpointController extends ControllerBase {

  /**
   * Kill-switch to avoid caching the page.
   *
   * @var \Drupal\Core\PageCache\ResponsePolicy\KillSwitch
   */
  private KillSwitch $cacheKillSwitch;

  /**
   * The discovery service.
   *
   * @var \Drupal\cl_components\Component\ComponentDiscovery
   */
  private ComponentDiscovery $componentDiscovery;

  /**
   * The renderer.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  private RendererInterface $renderer;

  /**
   * True if the inject syntax is supported.
   *
   * @var bool
   */
  private bool $supportsInject;

  /**
   * Creates an object.
   *
   * @param \Drupal\Core\PageCache\ResponsePolicy\KillSwitch $cache_kill_switch
   *   The cache kill switch.
   * @param \Drupal\cl_components\Component\ComponentDiscovery $component_discovery
   *   The component discovery.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface
   *   The module handler.
   */
  public function __construct(KillSwitch $cache_kill_switch, ComponentDiscovery $component_discovery, ModuleHandlerInterface $module_handler, RendererInterface $renderer) {
    $this->cacheKillSwitch = $cache_kill_switch;
    $this->componentDiscovery = $component_discovery;
    $this->supportsInject = $module_handler->moduleExists('cl_inject');
    $this->renderer = $renderer;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    $cache_kill_switch = $container->get('page_cache_kill_switch');
    assert($cache_kill_switch instanceof KillSwitch);
    $component_discovery = $container->get(ComponentDiscovery::class);
    assert($component_discovery instanceof ComponentDiscovery);
    $module_handler = \Drupal::service('module_handler');
    assert($module_handler instanceof ModuleHandlerInterface);
    $renderer = $container->get('renderer');
    return new static($cache_kill_switch, $component_discovery, $module_handler, $renderer);
  }

  /**
   * Render a Twig template from a Storybook component directory.
   */
  public function render(Request $request): array {
    $arguments = $this->getArguments($request);
    try {
      $component = $this->getComponent($request);
    }
    catch (ComponentNotFoundException $e) {
      throw new BadRequestHttpException('Invalid component', $e);
    }
    $variant = $request->query->get('_variant');
    // Storybook will not allow empty title for the story. Let's consider
    // Default to be the default.
    $variant = $variant === 'default' ? '' : $variant;
    try {
      $build = $this->generateRenderArray($component, $variant ?: '', $arguments);
    }
    catch (TemplateNotFoundException $e) {
      $build = [];
    }
    $this->cacheKillSwitch->trigger();
    return [
      '#attached' => [
        'library' => ['cl_components/attach_behaviors'],
      ],
      '#type' => 'container',
      '#cache' => ['max-age' => 0],
      // Magic wrapper ID to pull the HTML from.
      '#attributes' => ['id' => '___cl-wrapper'],
      'component' => $build,
    ];
  }

  /**
   * Gets the arguments.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The inbound request.
   *
   * @return array
   *   The array of arguments.
   */
  private function getArguments(Request $request): array {
    $params = $request->query->get('_params');
    $json = base64_decode($params, TRUE);
    if ($json === FALSE) {
      throw new BadRequestHttpException('Invalid component parameters');
    }
    return Json::decode($json);
  }

  /**
   * Get the component based on the request object.
   *
   * @throws \Drupal\cl_components\Exception\ComponentNotFoundException
   *   If the component cannot be found.
   */
  public function getComponent(Request $request): Component {
    $story_filename = $request->query->get('_storyFileName');
    if (!$story_filename) {
      throw new ComponentNotFoundException('Impossible to find a story with an empty story file name.');
    }
    $component = $this->componentDiscovery->findBySiblingFile($story_filename);
    if (!$component instanceof Component) {
      throw new ComponentNotFoundException('Impossible to find a component associated to the story. Make sure the story is in the same folder or in a sub0folder as the metadata.json file. Also check that metadata.json is valid.');
    }
    return $component;
  }

  /**
   * Generates a render array to showcase the component with the expected blocks.
   *
   * @param \Drupal\cl_components\Component\Component $component
   *   The component.
   * @param string $variant
   *   The variant.
   * @param array $context
   *   The template context.
   *
   * @return array
   *   The generated render array.
   *
   * @throws \Drupal\cl_components\Exception\TemplateNotFoundException
   */
  private function generateRenderArray(Component $component, string $variant, array $context): array {
    $metadata = $component->getMetadata();
    $component_schema = $metadata->getSchemas()['named_blocks'][$variant] ?? [];
    $block_names = array_keys($component_schema['properties'] ?? []);
    // If 'inject' is supported we should use it, unless there are block definitions in the template.
    if (empty($block_names) && $this->supportsInject) {
      $template = $this->generateInjectTemplate($component, $variant, $context);
      return [
        '#type' => 'inline_template',
        '#template' => $template,
        '#context' => $context,
      ];
    }
    return [
      '#type' => 'cl_component',
      '#component' => $component->getId(),
      '#variant' => $variant,
      '#twig_blocks' => array_intersect_key($context, array_flip($block_names)),
      '#trusted_blocks' => TRUE,
      '#context' => array_diff_key($context, array_flip($block_names)),
    ];
  }

  /**
   * Generates an 'inject' template for rendering a component in isolation.
   *
   * @param \Drupal\cl_components\Component\Component $component
   *   The component to render.
   * @param string $variant
   *   The component variant.
   * @param array $context
   *   The template context.
   *
   * @return string
   *   The template.
   */
  private function generateInjectTemplate(Component $component, string $variant, array $context): string {
    $component_id = $component->getId();
    $template = '{# This template was dynamically generated by cl_server #}' . PHP_EOL;
    $children = $context['children'] ?? '';
    $with = Json::encode(array_diff_key($context, array_flip(['children'])));
    $template .= empty($variant)
      ? '{% inject \'' . $component_id . '\' with ' . $with . ' %}'
      : '{% inject \'' . $component_id . '\' variant \'' . $variant . '\' with ' . $with . ' %}';

    $template .= $children;
    $template .= PHP_EOL . '{% endinject \'' . $component_id . '\' %}' . PHP_EOL;
    return $template;
  }

}
